# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import getpass
import itertools
import os
import re
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa
import structlog
import yaml

from pydiverse.common.util import deep_merge

if TYPE_CHECKING:
    from pydiverse.pipedag.context import ConfigContext


# noinspection PyPep8Naming
class cached_class_property:
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls):
        if not hasattr(self, "cache"):
            self.cache = self.func(cls)
        return self.cache


class PipedagConfig:
    """
    This class represents a :doc:`pipedag config file </reference/config>`.

    :param path: Path to the config yaml file to load or a dictionary containing the
        raw config as it would be loaded from yaml file.

    Attributes
    ----------
    default : PipedagConfig
        The default config file.

        If the environment variable :envvar:`PIPEDAG_CONFIG` is set, then this file
        will be used as the config file. Otherwise, pipedag searches for a file called
        ``pipedag.yaml`` or ``pipedag.yml`` in:

        * The current working directory
        * Any parent directories of the working directory
        * The user folder
    """

    default: "PipedagConfig"

    def __init__(self, path: str | Path | dict[str, Any]):
        self.path = None
        if isinstance(path, dict):
            self.raw_config = path
        else:
            if isinstance(path, str):
                path = Path(path)
            self.path = path
            with open(path) as f:
                self.raw_config = yaml.safe_load(f)

        self.config_dict = self.__parse_config(self.raw_config)

    @cached_class_property
    def default(cls):
        config_path = find_config()
        return PipedagConfig(config_path)

    @property
    def name(self) -> str:
        return self.config_dict.get("name")

    def __parse_config(self, raw_config: dict[str:Any]):
        config = copy.deepcopy(raw_config)

        # Expand all references
        for instance_dict in _get(config, "instances", default={}).values():
            self.__expand_references(instance_dict)

        for flow_dict in _get(config, "flows", default={}).values():
            self.__expand_references(flow_dict)
            for instance_dict in _get(flow_dict, "instances", default={}).values():
                self.__expand_references(instance_dict)

        return config

    def __expand_references(self, config):
        references = [
            (None, "technical_setup", "technical_setups"),
            ("table_store", "table_store_connection", "table_store_connections"),
            ("blob_store", "blob_store_connection", "blob_store_connections"),
        ]

        for expand_path, ref_name_path, ref_src_path in references:
            ref_name = _get(config, expand_path, ref_name_path, default=None)
            if ref_name is None:
                continue

            base_dict = _get(config, expand_path)
            merged_dict = deep_merge(base_dict, copy.deepcopy(_get(self.raw_config, ref_src_path, ref_name)))
            base_dict.update(merged_dict)
            _pop(base_dict, ref_name_path)

        return config

    def get(
        self,
        instance: str | None = None,
        flow: str | None = None,
        per_user: bool = False,
    ) -> "ConfigContext":
        """
        Constructs a :py:class:`ConfigContext`.
        For more details how the specific ConfigContext instance is constructed,
        check out the :ref:`specifying instances and flows
        <reference/config:Specifying instances and flows>` section.

        :param instance: Name of the instance.
            If no value is provided the ``__any__`` instance gets used.
        :param flow: Name of the flow.
            If no value is provided the ``__any__`` flow gets used.
        :param per_user:
            Whether to customize the instance id for each user according to
            :ref:`per_user_template`.
        """

        # TODO: Check that this function only gets called in the main interpreter.
        #       Otherwise certain environment variables might get expanded incorrectly.
        from pydiverse.pipedag.context import ConfigContext, default_config_dict

        config = self.__get_merged_config_dict(
            instance=instance,
            flow=flow,
            default=default_config_dict,
        ).copy()

        # TODO: Delegate selecting where variables can be expanded to the
        #  corresponding classes.
        #    eg. SQLTableStore._expand_env_vars = ["url", "url_attrs_file"]
        #    eg. SQLTableStore._expand_vars = ["url", "schema_prefix", "schema_suffix"]

        # First expand all environment variables
        self.__expand_environment_variables(inout_config=config)

        # Intermediate variable processing
        if per_user:
            template = config["per_user_template"]
            config["instance_id"] = expand_variables(
                template,
                {
                    "username": getpass.getuser(),
                    "id": config["instance_id"],
                },
            )

        # Handle url_attrs_file
        url_attrs_file = _pop(config, "table_store", "args", "url_attrs_file", default=None)
        if url_attrs_file is not None:
            with open(url_attrs_file, encoding="utf-8") as fh:
                url_attrs = yaml.safe_load(fh)

            url = _get(config, "table_store", "args", "url")
            url = expand_variables(url, url_attrs, skip_missing=True)
            _set(config, url, "table_store", "args", "url")

        # Finally, expand all normal variables
        config = self.__expand_variables(config)

        config_context = ConfigContext.new(config, self.name, flow, instance)

        if "PYDIVERSE_PIPEDAG_PYTEST" not in os.environ:
            # If we're running test cases, this can be skipped to improve performance
            try:
                # Make sure @cached_property store is set up and loaded
                # and throw config errors early.
                with config_context:
                    _ = config_context.store
                    _ = config_context.auto_table
                    _ = config_context.auto_blob

                    config_context.create_orchestration_engine().dispose()
                    config_context.create_lock_manager().dispose()
            except Exception as e:
                raise RuntimeError(
                    "Error while creating backend objects from pipedag config "
                    f"(instance={instance}, flow={flow}): {self.path}"
                ) from e

        return config_context

    def __get_merged_config_dict(self, instance, flow, default=None):
        search_paths = [
            ("instances", "__any__"),
            ("instances", instance),
            ("flows", "__any__"),
            ("flows", "__any__", "instances", instance),
            ("flows", flow),
            ("flows", flow, "instances", instance),
        ]

        search_paths = [path for path in search_paths if None not in path]
        dicts = [_get(self.config_dict, path, default=None) for path in search_paths]

        # Check for strict instance lookup
        # If instance is specified, make sure that a corresponding section can be found
        strict_instance_lookup = self.config_dict.get("strict_instance_lookup", True)
        if strict_instance_lookup and instance is not None:
            found_instance = False
            for path, d in zip(search_paths, dicts):
                found_instance |= "instances" in path and instance in path and d is not None

            if not found_instance:
                raise AttributeError(
                    f"Strict instance lookup failed: Couldn't find instance '{instance}' in pipedag config."
                )

        # Merge
        merged = default or {}
        for d in dicts:
            if d is not None:
                merged = deep_merge(merged, d)

        return merged

    @staticmethod
    def __expand_environment_variables(*, inout_config):
        locations = [
            ("table_store", "args", "url"),
            ("table_store", "args", "url_attrs_file"),
        ]

        for location in locations:
            value: str = _get(inout_config, location, default=None)
            if value is None:
                continue

            value = expand_environment_variables(value)
            _set(inout_config, value, location)

    def __expand_variables(self, config) -> dict[str, Any]:
        out_config = copy.deepcopy(config)
        locations = [
            ("table_store", "args", "url"),
            ("table_store", "args", "schema_prefix"),
            ("table_store", "args", "schema_suffix"),
        ]

        # TODO: Decide on a list of available variables
        variables = {
            "username": getpass.getuser(),
            "instance_id": config.get("instance_id"),
            "name": self.name,
        }

        for key, val in list(variables.items()):
            if val is None:
                variables.pop(key)

        for location in locations:
            value: str = _get(config, location, default=None)
            if value is None:
                continue

            value = expand_variables(value, variables)
            _set(out_config, value, location)
        return out_config


def create_basic_pipedag_config(
    engine_url: str,
    blob_directory: str | Path | None = None,
    name: str = "pipeline",
    instance_id: str = "pipeline",
    network_interface="127.0.0.1",
    kroki_url=None,
    auto_table=(
        "pandas.DataFrame",
        "sqlalchemy.sql.expression.TextClause",
        "sqlalchemy.sql.expression.Selectable",
    ),
    fail_fast=True,
    disable_stage_locking=False,
    file_locking_directory: str | Path | None = None,
) -> PipedagConfig:
    """
    Get code-based pipedag config as easy as providing a SQLAlchemy URL.

    This is an alternative to ``pipedag.yaml`` files which are typically loaded with
    ``PipedagConfig.default`` or ``PipedagConfig("path/pipedag.yaml")``.
    Feel free to extract ``create_basic_pipedag_config(url).raw_config``, modify it
    and to create a new ``PipedagConfig(raw_config)``.

    For more mature projects that configure multiple pipeline instances, it is
    recommended to stay with ``pipedag.yaml`` files though.

    Example::

        cfg = create_basic_pipedag_config(
            "duckdb:////tmp/pipedag/{instance_id}/db.duckdb",
            disable_stage_locking=True
        ).get("default")
        with cfg:
            with Flow() as flow:
                with Stage("step00"):
                    x = some_task()
            flow.run()

    Examples for engine_urls::

      "postgresql://{$DB_USER}:{$DB_PASSWORD}@{$DB_HOST}:{$DB_PORT}/{instance_id}"
      "postgresql://sa:Pydiverse23@127.0.0.1:6543/{instance_id}"
      "mssql+pyodbc://sa:PydiQuant27@127.0.0.1:1433/{instance_id}?driver=ODBC+Driver+18+for+SQL+Server&encrypt=no"
      "db2+ibm_db://db2inst1:password@localhost:50000/testdb"
      "duckdb:////tmp/pipedag/{instance_id}/db.duckdb"

    :param engine_url:
        SQLAlchemy engine url. It may include references to ENVIRONMENT VARIABLES
        like ``{$DB_PASSWORD}``. Furthermore, ``{name}`` is replaced with pipeline
        name and ``{instance_id}`` is replaced with the instance_id (see below).
    :param blob_directory:
        Default None;
        Directory where blobs are stored. In case of None, no task in the pipeline
        is allowed to return a ``Blob()`` object.
    :param name: pipeline name
    :param instance_id: ``instance_id`` used as placeholder in parameters ``engine_url``
        and ``blob_directory``. It is more relevant for non-basic configuration with
        multiple pipeline instances.
    :param network_interface:
        Used for shared state communication while executing flow. It is possible
        to perform multi-node parallel execution with the DaskEngine. You probably
        don't need to worry and just leave it at ``127.0.0.1`` for localhost
    :param kroki_url:
        Default None; Sets the URL used for creating links after each flow execution
        which allow you to visualize the pipeline run. When setting the URL here,
        it will automatically set disable_kroki to False.
    :param auto_table:
        You might be just happy with the default to convert pandas DataFrames
        and SQLAlchemy clauses to Tables when returning them in tasks. If you use
        pipedag.Table() objects for manually choosing table names, this is not
        relevant.
    :param fail_fast:
        Default True;
        Aborts execution of pipeline on first error. This is often easier to
        read stacktraces. If you want to see all errors a more complex DAG, please
        set to False.
    :param disable_stage_locking:
        Default False;
        When True, choose NoLockManager as opposed to DatabaseLockManager. In this
        case, two parallel runs of the pipeline can easily corrupt schema swapping.
        Unfortunately, we don't support database locking for duckdb engine_urls.
        So you must either use ``disable_stage_locking=True`` or set
        ``file_locking_directory``.
    :param file_locking_directory:
        Default None;
        If string or Path is provided, chose file based locking in the given directory.
        We use database locking by default (``file_locking_directory=None`` and
        ``disable_stage_locking=False``).
        However, we only support this for MSSQL, Postgres, and DB2 dialects so far.
    :return:
        PipedagConfig object
    """

    # test engine:
    engine = sa.create_engine(engine_url)
    dialect = engine.dialect.name
    engine.dispose()

    if dialect not in ["mssql", "postgresql", "ibm_db_sa", "duckdb"]:
        logger = structlog.get_logger(logger_name=__name__, function="create_basic_pipedag_config")
        logger.info("")

    if (
        dialect not in ["mssql", "postgresql", "ibm_db_sa"]
        and not disable_stage_locking
        and file_locking_directory is None
    ):
        raise AttributeError(
            f"We don't support Database Locking for database dialect {dialect}. "
            "Please consider setting disable_stage_locking=True (i.e. for local DuckDB)"
            " or file_locking_directory for team shared database."
        )

    if dialect in ["duckdb", "ibm_db_sa"]:
        stage_commit_technique = "read_views"
    else:
        stage_commit_technique = "schema_swap"
    table_store = {
        "class": "pydiverse.pipedag.backend.table.SQLTableStore",
        "args": dict(url=engine_url, print_materialize=True, print_sql=True),
    }
    if blob_directory is None:
        blob_store = {"class": "pydiverse.pipedag.backend.blob.NoBlobStore"}
    else:
        blob_store = {
            "class": "pydiverse.pipedag.backend.blob.FileBlobStore",
            "args": dict(base_path=str(blob_directory)),
        }
    if disable_stage_locking:
        lock_manager = {"class": "pydiverse.pipedag.backend.lock.NoLockManager"}
    elif file_locking_directory is not None:
        lock_manager = {
            "class": "pydiverse.pipedag.backend.lock.FileLockManager",
            "args": dict(base_path=str(file_locking_directory)),
        }
    else:
        lock_manager = {"class": "pydiverse.pipedag.backend.lock.DatabaseLockManager"}
    orchestration = {"class": "pydiverse.pipedag.engine.SequentialEngine"}
    raw_config = dict(
        name=name,
        strict_instance_lookup=False,
        instances=dict(
            __any__=dict(
                network_interface=network_interface,
                auto_table=list(auto_table),
                fail_fast=fail_fast,
                disable_kroki=kroki_url is None,
                kroki_url=kroki_url,
                instance_id=instance_id,
                stage_commit_technique=stage_commit_technique,
                table_store=table_store,
                blob_store=blob_store,
                lock_manager=lock_manager,
                orchestration=orchestration,
            )
        ),
    )
    return PipedagConfig(raw_config)


def find_config(
    name: str = "pipedag",
    search_paths: Iterable[str | Path] = None,
) -> str:
    """Searches for a pipedag config file

    The following paths get checked first.

    - The path specified in the "PIPEDAG_CONFIG" environment variable

    Else it searches in the following locations:

    - Current working directory
    - All parent directories
    - The user folder

    :param name: The name of the config file
    :param search_paths: The directories in which to search for the config file
    :return: The path of the file.
    :raises FileNotFoundError: if no config file could be found.
    """

    extensions = [".yaml", ".yml"]

    if search_paths is None:
        # Check PIPEDAG_CONFIG path
        if path := os.environ.get("PIPEDAG_CONFIG", None):
            path = Path(path).resolve().expanduser()
            if path.is_file():
                return str(path)

            for extension in extensions:
                path = path / (name + extension)
                if path.is_file():
                    return str(path)

        # Else, search in these default directories
        search_paths = [
            Path.cwd(),
            *Path.cwd().resolve().parents,
            Path("~").expanduser(),
        ]
    else:
        search_paths = [Path(path) for path in search_paths]

    file_names = [name + extension for extension in extensions]
    for path, file_name in itertools.product(search_paths, file_names):
        config_path: Path = (path / file_name).resolve()
        if config_path.is_file():
            return str(config_path)

    raise FileNotFoundError("No config file found")


def expand_environment_variables(string: str) -> str:
    """
    Expands all occurrences of the form ``{$ENV_VAR}`` with the environment variable
    named ``ENV_VAR``.
    """

    def env_var_sub(match: re.Match):
        name = match.group()[2:-1]
        if name not in os.environ:
            raise AttributeError(f"Could not find environment variable '{name}' referenced in '{string}'.")
        return os.environ[name]

    return re.sub(r"\{\$[a-zA-Z_]+[a-zA-Z0-9_]*\}", env_var_sub, string)


def expand_variables(string: str, variables: dict[str:str], skip_missing: bool = False) -> str:
    """
    Expands all occurrences of the form {var_name} with the variable
    named ``var_name``.
    """

    def var_sub(match: re.Match):
        name = match.group()[1:-1]
        if name not in variables:
            if skip_missing:
                return match.group()
            raise AttributeError(f"Could not find variable '{name}' referenced in '{string}'.")
        return str(variables[name])

    return re.sub(r"\{[a-zA-Z_]+[a-zA-Z0-9_]*\}", var_sub, string)


# Nested Dictionary Utilities


def _flatten(items):
    for x in items:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from _flatten(x)
        else:
            yield x


_nil = object()


def _get(d, *path, default=_nil):
    path = [x for x in _flatten(path) if x is not None]
    try:
        for p in path:
            if p is not None:
                d = d[p]
    except (IndexError, KeyError) as e:
        if default is _nil:
            raise KeyError(f"Invalid path: {' > '.join(path)}") from e
        return default

    return d


def _set(d, value, *path):
    path = [x for x in _flatten(path) if x is not None]
    for p in path[:-1]:
        d = d[p]
    d[path[-1]] = value


def _pop(d, *path, default=_nil) -> Any:
    try:
        path = [x for x in _flatten(path) if x is not None]
        for p in path[:-1]:
            d = d[p]
        return d.pop(path[-1])
    except KeyError as e:
        if default == _nil:
            raise e
        return default



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import importlib
import warnings
from pathlib import Path
from typing import Any

import pandas as pd
import sqlalchemy as sa

from pydiverse.common import Dtype
from pydiverse.pipedag import Table
from pydiverse.pipedag.backend.table.sql.hooks import (
    IbisTableHook,
    PandasTableHook,
    PolarsTableHook,
)
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore
from pydiverse.pipedag.container import Schema
from pydiverse.pipedag.materialize.details import resolve_materialization_details_label

try:
    import duckdb
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    duckdb = None


class DuckDBTableStore(SQLTableStore):
    """
    SQLTableStore that supports `DuckDB <https://duckdb.org>`_.

    Takes the same arguments as
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`
    """

    _dialect_name = "duckdb"

    def _metadata_pk(self, name: str, table_name: str):
        sequence = sa.Sequence(f"{table_name}_{name}_seq", metadata=self.sql_metadata)
        return sa.Column(
            name,
            sa.BigInteger(),
            sequence,
            server_default=sequence.next_value(),
            primary_key=True,
        )

    def _default_isolation_level(self) -> str | None:
        return None  # "READ UNCOMMITTED" does not exist in DuckDB

    def _init_database_before_engine(self):
        # Duckdb already creates the database file automatically
        # However, the parent directory doesn't get created automatically
        database = self.engine_url.database
        if database is None or database == ":memory:":
            return

        database_path = Path(database)
        database_path.parent.mkdir(parents=True, exist_ok=True)

    def _init_database(self):
        if not self.create_database_if_not_exists:
            return

    def dialect_requests_empty_creation(self, table: Table, is_sql: bool) -> bool:
        _ = table, is_sql
        return False  # DuckDB is not good with stable type arithmetic


@DuckDBTableStore.register_table(pd)
class PandasTableHook(PandasTableHook):
    @classmethod
    def _execute_materialize(
        cls,
        table: Table[pd.DataFrame],
        store: DuckDBTableStore,
        schema: Schema,
        dtypes: dict[str, Dtype],
    ):
        df = table.obj
        engine = store.engine
        dtypes = cls._get_dialect_dtypes(dtypes, table)

        store.check_materialization_details_supported(resolve_materialization_details_label(table))

        # Create empty table with correct schema
        cls._dialect_create_empty_table(store, table, schema, dtypes)
        store.add_indexes_and_set_nullable(table, schema, on_empty_table=True, table_cols=df.columns)

        # Copy dataframe directly to duckdb
        # This is SIGNIFICANTLY faster than using pandas.to_sql
        table_name = engine.dialect.identifier_preparer.quote(table.name)
        schema_name = engine.dialect.identifier_preparer.format_schema(schema.get())

        conn = engine.raw_connection()
        # Attention: This sql copies local variable df into database (FROM df)
        conn.execute(f"INSERT INTO {schema_name}.{table_name} SELECT * FROM df")

        store.add_indexes_and_set_nullable(
            table,
            schema,
            on_empty_table=False,
            table_cols=df.columns,
        )


try:
    import polars as pl
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    pl = importlib.import_module("polars")
    pl.DataType = None
    pl.DataFrame = None
    pl.LazyFrame = None


@DuckDBTableStore.register_table(pl.DataFrame, duckdb)
class PolarsTableHook(PolarsTableHook):
    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) does not support DuckDB.
        return False

    @classmethod
    def download_table(
        cls,
        query: Any,
        store: SQLTableStore,
        dtypes: dict[str, pl.DataType] | None = None,
    ) -> pl.DataFrame:
        assert dtypes is None, (
            "Polars reads SQL schema and loads the data in reasonable types."
            "Thus, manual dtype manipulation can only done via query or afterwards."
        )
        engine = store.engine
        # Connectorx doesn't support duckdb.
        # Instead, we load it like this:  DuckDB -> PyArrow -> Polars
        conn = engine.raw_connection()
        pl_table = conn.sql(query).arrow()

        df = pl.from_arrow(pl_table)
        return df


try:
    import ibis
except ImportError:
    ibis = None


@DuckDBTableStore.register_table(ibis)
class IbisTableHook(IbisTableHook):
    @classmethod
    def _conn(cls, store: DuckDBTableStore):
        return ibis.duckdb.from_connection(store.engine.raw_connection())



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pandas as pd
import sqlalchemy as sa

from pydiverse.pipedag import Blob, Flow, Stage, Table
from pydiverse.pipedag.container import RawSql
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.context.context import CacheValidationMode, ConfigContext
from pydiverse.pipedag.materialize.core import materialize

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import ALL_INSTANCES, skip_instances, with_instances
from tests.util import select_as
from tests.util import tasks_library as m
from tests.util.spy import spy_task

# snowflake tests are too slow, possibly they could move to nightly tests
pytestmark = [with_instances(tuple(set(ALL_INSTANCES) - {"snowflake"}))]


# Test that running a flow that contains a task with an invalid cache function
# doesn't trigger that task when run with ignore_cache_function=True, and it is
# otherwise cache valid.


def test_literal(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_value():
        return cache_value

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_value()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value while setting ignore_cache_function=True should
    # ignore the cache function.
    cache_value = 1
    for _ in range(3):
        with StageLockContext():
            result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
            assert result.get(out) == 0
            assert result.get(child) == 0
            out_spy.assert_not_called()
            child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_table(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_table():
        return Table(select_as(cache_value, "x"))

    @materialize(input_type=pd.DataFrame, version="1.0")
    def get_first(table, col):
        return int(table[col][0])

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_table()
            child = get_first(out, "x")

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value while setting ignore_cache_function=True should
    # ignore the cache function.
    cache_value = 1
    for _ in range(3):
        with StageLockContext():
            result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
            assert result.get(child) == 0
            out_spy.assert_not_called()
            child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_lazy_table(mocker):
    cache_value = 0
    lazy_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, lazy=True, nout=2)
    def input_task():
        return Table(select_as(lazy_value, "x")), cache_value

    @materialize(input_type=pd.DataFrame, version="1.0")
    def get_first(table, col):
        return int(table[col][0])

    with Flow() as flow:
        with Stage("stage_1"):
            out, out_cache = input_task()
            child = get_first(out, "x")
            cache_child = m.noop(out_cache)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 0

    # Calling flow.run again should call the lazy task but not the child task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    cache_spy = spy_task(mocker, cache_child)
    for _ in range(3):
        with StageLockContext():
            result = flow.run()
            assert result.get(child) == 0
            assert result.get(cache_child) == 0
            out_spy.assert_called_once()
            child_spy.assert_not_called()
            cache_spy.assert_not_called()

    # Changing the cache value while setting ignore_cache_function=True should
    # still call the lazy task.
    cache_value = 1
    with StageLockContext():
        result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
        assert result.get(child) == 0
        assert result.get(cache_child) == 1
        out_spy.assert_called_once()
        child_spy.assert_not_called()
        cache_spy.assert_called_once()

    # Only changing the lazy_value should cause the child task to get called
    lazy_value = 1
    with StageLockContext():
        result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
        assert result.get(child) == 1
        assert result.get(cache_child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()
        cache_spy.assert_not_called()

    # We can't avoid the cache invalidation of get_first here since we do
    # ignore_cache_function based cache_fn_hash filtering on task level and
    # not on lazy table level.
    with StageLockContext():
        result = flow.run()
        assert result.get(child) == 1
        assert result.get(cache_child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()
        cache_spy.assert_not_called()

    # The child task shouldn't get called again, because the lazy sql didn't change
    for _ in range(3):
        with StageLockContext():
            result = flow.run()
            assert result.get(child) == 1
            assert result.get(cache_child) == 1
            out_spy.assert_called_once()
            child_spy.assert_not_called()
            cache_spy.assert_not_called()


@with_instances("postgres")
def test_blob(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_blob():
        return Blob(cache_value)

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_blob()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value while setting ignore_cache_function=True should
    # ignore the cache function.
    cache_value = 1
    for _ in range(3):
        with StageLockContext():
            result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
            assert result.get(out) == 0
            assert result.get(child) == 0
            out_spy.assert_not_called()
            child_spy.assert_not_called()

    # Changing the cache value should cause it to get called again
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def ref(tbl):
    return f"{tbl.original.schema}.{tbl.original.name}"


# TODO: Determine exactly why this test only works with postgres
@skip_instances(
    "mssql",
    "mssql_pytsql",
    "mssql_nobulk",
    "mssql_noaodbc",
    "mssql_columnstore",
    "ibm_db2",
    "ibm_db2_avoid_schema",
    "ibm_db2_materialization_details",
    "duckdb",
)
def test_raw_sql(mocker):
    cache_value = 0
    raw_value = 0

    def cache(*args, **kwargs):
        return cache_value

    @materialize(lazy=True, cache=cache)
    def raw_sql_task(stage):
        store = ConfigContext.get().store.table_store
        schema = store.get_schema(stage.transaction_name).get()
        return RawSql(
            f"""
            CREATE TABLE {schema}.raw_table AS
            SELECT {raw_value} as x
            """
        )

    @materialize(version="1.0", input_type=sa.Table)
    def child_task(raw_sql):
        return Table(sa.text(f"SELECT * FROM {ref(raw_sql['raw_table'])}"))

    with Flow() as flow:
        with Stage("raw_sql_stage") as stage:
            out = raw_sql_task(stage)
            child = child_task(out)

    # Initial Run
    with StageLockContext():
        result = flow.run()
        assert result.get(child, as_type=pd.DataFrame)["x"][0] == 0

    # Calling flow.run again should call the raw sql task but not the child task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(child, as_type=pd.DataFrame)["x"][0] == 0
        out_spy.assert_called_once()
        child_spy.assert_not_called()

    # Changing the cache value while setting ignore_cache_function=True should
    # still call the raw sql task.
    cache_value = 1
    for _ in range(3):
        with StageLockContext():
            result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
            assert result.get(child, as_type=pd.DataFrame)["x"][0] == 0
            out_spy.assert_called_once()
            child_spy.assert_not_called()

    # Only changing the raw_value should cause the child task to get called
    raw_value = 1
    with StageLockContext():
        result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
        assert result.get(child, as_type=pd.DataFrame)["x"][0] == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()

    # We can't avoid the cache invalidation of get_first here since we do
    # ignore_cache_function based cache_fn_hash filtering on task level and
    # not on lazy table level.
    with StageLockContext():
        result = flow.run()
        assert result.get(child, as_type=pd.DataFrame)["x"][0] == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()

    # The child task shouldn't get called again, because the raw sql didn't change
    for _ in range(3):
        with StageLockContext():
            result = flow.run()
            assert result.get(child, as_type=pd.DataFrame)["x"][0] == 1
            out_spy.assert_called_once()
            child_spy.assert_not_called()


# Some more complicated tests that validate the behaviour of
# ignore_cache_function=True


def test_input_invalid(mocker):
    # Test that it does get run if it is otherwise cache invalid
    lazy_value = 0

    @materialize(lazy=True)
    def input_task():
        return lazy_value

    with Flow() as flow:
        with Stage("stage_1"):
            out = input_task()
            child = m.noop(out)

    with StageLockContext():
        result = flow.run()
        assert result.successful
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Setting ignore_cache_function shouldn't have an influence on a lazy task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_called_once()
        child_spy.assert_not_called()

    # Despite ignore_cache_function=True, the child tasks should still run
    # because its inputs changed
    lazy_value = 1
    with StageLockContext():
        result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
        assert result.get(out) == 1
        assert result.get(child) == 1
        out_spy.assert_called_once()
        child_spy.assert_called_once()


def test_cache_temporarily_different(mocker):
    cache_value = 0

    def cache():
        return cache_value

    @materialize(cache=cache, version="1.0")
    def return_cache_value():
        return cache_value

    with Flow() as flow:
        with Stage("stage_1"):
            out = return_cache_value()
            child = m.noop(out)

    # Initial Call
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0

    # Calling flow.run again shouldn't call the task
    out_spy = spy_task(mocker, out)
    child_spy = spy_task(mocker, child)
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()

    # Changing the cache value while setting ignore_cache_function=True should
    # ignore the cache function.
    cache_value = 1
    for _ in range(3):
        with StageLockContext():
            result = flow.run(cache_validation_mode=CacheValidationMode.IGNORE_FRESH_INPUT)
            assert result.get(out) == 0
            assert result.get(child) == 0
            out_spy.assert_not_called()
            child_spy.assert_not_called()

    # Changing the cache value back to the original shouldn't trigger
    # any tasks to run
    cache_value = 0
    with StageLockContext():
        result = flow.run()
        assert result.get(out) == 0
        assert result.get(child) == 0
        out_spy.assert_not_called()
        child_spy.assert_not_called()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from concurrent.futures.process import BrokenProcessPool

import pandas as pd
import pytest
import sqlalchemy as sa
import structlog

from pydiverse.pipedag import ConfigContext, Flow, Stage, Table, materialize
from pydiverse.pipedag.context import FinalTaskState, RunContext, StageLockContext
from pydiverse.pipedag.context.context import CacheValidationMode
from pydiverse.pipedag.context.trace_hook import PrintTraceHook
from pydiverse.pipedag.core.config import PipedagConfig

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import (
    ALL_INSTANCES,
    ORCHESTRATION_INSTANCES,
    skip_instances,
    with_instances,
)
from tests.util import select_as, swallowing_raises
from tests.util import tasks_library as m
from tests.util import tasks_library_imperative as m2

pytestmark = [with_instances(ALL_INSTANCES, ORCHESTRATION_INSTANCES)]


def test_materialize_literals():
    """
    Check that all json serializable literals get stored and retrieved
    correctly from the table / metadata store.
    """

    def materialize_and_retrieve(value):
        v = m.noop(value)
        return m.assert_equal(value, v)

    with Flow("flow") as f:
        with Stage("stage"):
            materialize_and_retrieve(None)

            materialize_and_retrieve(0)
            materialize_and_retrieve(1)

            # max 32bit
            materialize_and_retrieve(-2147483648)
            materialize_and_retrieve(2147483647)

            # max 64bit
            materialize_and_retrieve(-9223372036854775808)
            materialize_and_retrieve(9223372036854775807)

            # in python json we even have infinite precision integers:
            materialize_and_retrieve(-9223372036854775809)
            materialize_and_retrieve(9223372036854775808)

            materialize_and_retrieve(3.14)
            materialize_and_retrieve(-2.71)

            materialize_and_retrieve("")
            materialize_and_retrieve("a string")
            materialize_and_retrieve("a" * 256)

            materialize_and_retrieve([0, 1, 2, 3])

            materialize_and_retrieve({"0": 0, "1": 1})

            materialize_and_retrieve(
                {
                    "key": [0, 1, 2, 3],
                    "numbers": {
                        "zero": 0,
                        "one": 1,
                    },
                    "nested_lists": [[[None]]],
                }
            )

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            _m.assert_table_equal(x, x)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table_subtask(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_subtask()
            y = _m.noop_subtask(x)
            z = _m.noop_subtask_lazy(x)
            _m.assert_table_equal(x, y)
            _m.assert_table_equal(x, z)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table_subtask_fail_input_type(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_subtask()
            _ = _m.noop_subtask_fail_input_type(x)
    with pytest.raises(RuntimeError, match="does not match parent task input type"):
        try:
            f.run()
        except BrokenProcessPool as e:
            # TODO: find out whether we can avoid BrokenProcessPool from happening
            raise RuntimeError("does not match parent task input type + BrokenProcessPool") from e


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_table_twice(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe()
            y = _m.simple_dataframe()

            m.assert_table_equal(x, y)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_debug_materialize_table_no_taint(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_debug_materialize_no_taint()
            _m.assert_table_equal(x, x)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_debug_materialize_table_twice(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage"):
            x = _m.simple_dataframe_debug_materialize_twice()
            _m.assert_table_equal(x, x)

    with pytest.raises(RuntimeError, match="interactive debugging"):
        assert f.run().successful


@with_instances("postgres")
def test_imperative_minimal_example():
    with Flow("flow") as f:
        with Stage("input"):
            in_ = m.simple_dataframe()
        with Stage("stage"):
            _ = m.complex_imperative_materialize(in_)

    trace_hook = PrintTraceHook()  # can be used to debug caching issues
    result1 = f.run(
        cache_validation_mode=CacheValidationMode.FORCE_CACHE_INVALID,
        trace_hook=trace_hook,
    )

    result2 = f.run(trace_hook=trace_hook)

    print("Run 1:")
    for task in f.tasks:
        print(f"{task.name}: {result1.task_states.get(task)}")

    print("Run 2:")
    for task in f.tasks:
        print(f"{task.name}: {result2.task_states.get(task)}")

    # Expectation: In Run 2 all tasks should be cached
    for task in f.tasks:
        if not task.name.startswith("Commit"):
            assert result2.task_states.get(task) == FinalTaskState.CACHE_VALID


@with_instances("postgres")
def test_materialize_blob():
    with Flow("flow") as f:
        with Stage("stage_0"):
            x = m.object_blob({"a": 12, "b": 24})
            y = m.as_blob(x)
            m.assert_blob_equal(x, y)

        with Stage("stage_1"):
            one = m.one()
            one_blob = m.as_blob(m.one())
            m.assert_equal(one, one_blob)

    assert f.run().successful


def test_failure():
    with Flow("flow") as f:
        with Stage("failure_stage"):
            m.exception(0, True)

    with swallowing_raises(Exception, match="THIS EXCEPTION IS EXPECTED"):
        f.run(fail_fast=True)

    with Flow("flow") as f:
        with Stage("failure_stage"):
            x = m.exception(0, True)
            m.noop(x)

    with swallowing_raises(Exception, match="THIS EXCEPTION IS EXPECTED"):
        f.run(fail_fast=True)


def test_return_pipedag_config():
    with Flow("flow") as f:
        with Stage("failure_stage"):
            m.noop(PipedagConfig.default)

    with swallowing_raises(TypeError, match="PipedagConfig"):
        f.run(fail_fast=True)


def test_return_config_context():
    with Flow("flow") as f:
        with Stage("failure_stage"):
            m.noop(PipedagConfig.default.get())

    with swallowing_raises(TypeError, match="ConfigContext"):
        f.run(fail_fast=True)


def test_materialize_memo_literal():
    # A flow should be able to contain the same task with the same inputs
    # more than once and still run successfully.

    with Flow("flow") as f:
        with Stage("stage_0"):
            t_0 = m.noop(1)
            t_1 = m.noop(1)

            one = m.one()
            t_2 = m.noop(one)
            t_3 = m.noop(one)

        with Stage("stage_1"):
            t_4 = m.noop(1)
            t_5 = m.noop(t_4)
            t_6 = m.noop(t_5)
            t_7 = m.noop(t_0)

        with Stage("stage_2"):
            m.assert_equal(t_0, t_4)
            m.assert_equal(t_1, t_5)
            m.assert_equal(t_2, t_6)
            m.assert_equal(t_3, t_7)

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_materialize_memo_table(imperative):
    _m = m if not imperative else m2
    with Flow("flow") as f:
        with Stage("stage_0"):
            t_0 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_1 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_2 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_3 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})

            _m.assert_table_equal(t_0, t_1)
            _m.assert_table_equal(t_1, t_2)
            _m.assert_table_equal(t_2, t_3)

        with Stage("stage_1"):
            t_4 = _m.pd_dataframe({"x": [0, 0], "y": [0, 0]})
            t_5 = _m.noop(t_4)
            t_6 = _m.noop(t_5)
            t_7 = _m.noop(t_0)
            t_8 = _m.noop_sql(t_4)
            t_9 = _m.noop_sql(t_5)
            t_10 = _m.noop_sql(t_0)
            t_11 = _m.noop_lazy(t_4)
            t_12 = _m.noop_lazy(t_5)
            t_13 = _m.noop_lazy(t_0)
            if _m.pl is not None:
                t_14 = _m.noop_polars(t_4)
                t_15 = _m.noop_polars(t_5)
                t_16 = _m.noop_polars(t_0)
                t_17 = _m.noop_lazy_polars(t_4)
                t_18 = _m.noop_lazy_polars(t_5)
                t_19 = _m.noop_lazy_polars(t_0)

        with Stage("stage_2"):
            _m.assert_table_equal(t_0, t_4)
            _m.assert_table_equal(t_1, t_5)
            _m.assert_table_equal(t_2, t_6)
            _m.assert_table_equal(t_3, t_7)
            _m.assert_table_equal(t_1, t_8)
            _m.assert_table_equal(t_2, t_9)
            _m.assert_table_equal(t_3, t_10)
            _m.assert_table_equal(t_1, t_11)
            _m.assert_table_equal(t_2, t_12)
            _m.assert_table_equal(t_3, t_13)
            if _m.pl is not None:
                _m.assert_table_equal(t_1, t_14)
                _m.assert_table_equal(t_2, t_15)
                _m.assert_table_equal(t_3, t_16)
                _m.assert_table_equal(t_1, t_17)
                _m.assert_table_equal(t_2, t_18)
                _m.assert_table_equal(t_3, t_19)

    assert f.run().successful


def test_materialize_memo_with_failure():
    with Flow("flow") as f:
        with Stage("stage"):
            m.exception(1, False)
            m.exception(1, False)

            one = m.one()
            m.exception(1, True)
            t_3 = m.exception(one, True)
            m.exception(t_3, False)

    with ConfigContext.get().evolve(swallow_exceptions=True):
        assert not f.run(fail_fast=False).successful


def test_stage_ref_counter():
    def check_rc(stage, expected):
        @materialize(lazy=True, input_type=pd.DataFrame)
        def _m_check_rc(x):
            assert RunContext.get().get_stage_ref_count(stage) == expected
            return x

        return _m_check_rc

    # Super simple case with just two task inside one stage
    with Flow("flow") as f:
        with Stage("stage") as s:
            task_1 = m.one()
            task_2 = m.two()

            # One reference from the check_rc and commit
            check_rc(s, 2)([task_1, task_2])

    assert f.run().successful

    # Multiple tasks with interdependency inside one stage
    with Flow("flow") as f:
        with Stage("stage") as s:
            task_1 = m.one()
            task_2 = m.two()
            task_tuple = m.noop((task_1, task_2))
            # One reference from assert, noop, assert and commit
            task_tuple = check_rc(s, 4)(task_tuple)
            task_tuple = m.noop(task_tuple)
            # One reference from assert and commit
            check_rc(s, 2)(task_tuple)

    assert f.run().successful

    # Multiple tasks spread over multiple stages
    with Flow("flow") as f:
        with Stage("stage_1") as s1:
            task_1 = m.one()
            # One reference from assert, noop, assert, commit and downstream
            task_1 = check_rc(s1, 5)(task_1)
            task_1 = m.noop(task_1)
            # One reference from assert, commit and downstream
            check_rc(s1, 3)(task_1)

        with Stage("stage_2") as s2:
            task_2 = m.two()
            task_2 = m.noop([task_2])
            check_rc(s2, 3)(task_2)

        with Stage("stage_3") as s3:
            task_tuple = m.noop((task_1, task_2))
            # Check that s1 and s2 have been freed
            x = check_rc(s1, 0)(task_tuple)
            x = check_rc(s2, 0)(x)
            check_rc(s3, 2)(x)

    assert f.run().successful


def test_nout_and_getitem():
    with Flow("flow") as f:
        with Stage("stage_0"):
            x = m.noop((0, 1))
            zero, one = m.create_tuple(x[0], x[1])

            m.assert_equal(zero, 0)
            m.assert_equal(one, m.one())

        with Stage("stage_1"):
            d = m.noop({"a": [0, 1, 2, 3], "b": [3, 2, 1, 0]})

            a = d["a"]
            b = d["b"]

            for i in range(4):
                m.assert_equal(a[i], i)
                m.assert_equal(a[i], b[3 - i])

    assert f.run().successful


@pytest.mark.parametrize("imperative", [False, True])
def test_name_mangling_tables(imperative):
    @materialize
    def table_task_1():
        tbl = Table(pd.DataFrame({"x": [1]}), name="table_%%")
        return tbl.materialize() if imperative else tbl

    @materialize
    def table_task_2():
        tbl = Table(pd.DataFrame({"x": [2]}), name="table_%%")
        return tbl.materialize() if imperative else tbl

    with Flow() as f:
        with Stage("stage_1"):
            table_1 = table_task_1()
            table_2 = table_task_2()

    with StageLockContext():
        result = f.run()
        assert result.get(table_1, as_type=pd.DataFrame)["x"][0] == 1
        assert result.get(table_2, as_type=pd.DataFrame)["x"][0] == 2


@pytest.mark.parametrize("imperative", [False, True])
def test_name_mangling_lazy_tables(imperative):
    @materialize(lazy=True)
    def lazy_task_1():
        tbl = Table(select_as(1, "x"), name="table_%%")
        return tbl.materialize() if imperative else tbl

    @materialize(lazy=True)
    def lazy_task_2():
        tbl = Table(select_as(2, "x"), name="table_%%")
        return tbl.materialize() if imperative else tbl

    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            lazy_2 = lazy_task_2()

    with StageLockContext():
        result = f.run()
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1
        assert result.get(lazy_2, as_type=pd.DataFrame)["x"][0] == 2


@pytest.mark.parametrize("imperative", [False, True])
def test_name_mangling_lazy_table_cache_fn(imperative):
    # Only the cache fn output of these two tasks is different
    @materialize(lazy=True, cache=lambda: 1, name="lazy_task")
    def lazy_task_1():
        tbl = Table(select_as(1, "x"))
        return tbl.materialize() if imperative else tbl

    @materialize(lazy=True, cache=lambda: 2, name="lazy_task")
    def lazy_task_2():
        tbl = Table(select_as(2, "x"))
        return tbl.materialize() if imperative else tbl

    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            lazy_2 = lazy_task_2()

    with StageLockContext():
        result = f.run()
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1
        assert result.get(lazy_2, as_type=pd.DataFrame)["x"][0] == 2


def test_run_flow_with_empty_stage():
    with Flow("f") as f:
        with Stage("stage_0"):
            pass

    assert f.run().successful


@materialize(lazy=True, input_type=sa.Table)
def _check_nullable_task(tables):
    assert len(tables) == 18
    # This is dialect specific:
    # assert [c.nullable for c in tables[0].c] == [True, True, True]
    assert [c.nullable for c in tables[1].c] == [False, True, False]
    assert [c.nullable for c in tables[2].c] == [True, False, True]
    assert [c.nullable for c in tables[3].c] == [True, False, True]
    assert [c.nullable for c in tables[4].c] == [False, True, False]
    assert [c.nullable for c in tables[5].c] == [True, True, True]
    assert [c.nullable for c in tables[6].c] == [True, False, True]
    assert [c.nullable for c in tables[7].c] == [False, True, False]
    assert [c.nullable for c in tables[8].c] == [False, False, False]
    # assert [c.nullable for c in tables[9].c] == [True, True, True]
    assert [c.nullable for c in tables[10].c][1:] == [True, False]
    assert [c.nullable for c in tables[11].c][1:] == [False, True]
    assert [c.nullable for c in tables[12].c][1:] == [False, True]
    assert [c.nullable for c in tables[13].c][1:] == [True, False]
    assert [c.nullable for c in tables[14].c][1:] == [True, True]
    assert [c.nullable for c in tables[15].c][1:] == [False, True]
    assert [c.nullable for c in tables[16].c][1:] == [True, False]
    assert [c.nullable for c in tables[17].c][1:] == [False, False]
    return tables


def _test_nullable_lazy_get_flow():
    cols = sa.literal(1).label("x"), sa.literal(2).label("y"), sa.literal(3).label("z")

    @materialize(lazy=True)
    def lazy_task_1():
        q = sa.select(*cols)
        tables = [
            Table(q),
            Table(q, nullable=["y"]),
            Table(q, nullable=["x", "z"]),
            Table(q, non_nullable=["y"]),
            Table(q, non_nullable=["x", "z"]),
            Table(q, nullable=["x", "y", "z"], non_nullable=[]),
            Table(q, nullable=["x", "z"], non_nullable=["y"]),
            Table(q, nullable=["y"], non_nullable=["x", "z"]),
            Table(q, nullable=[], non_nullable=["x", "y", "z"]),
            Table(q, primary_key=["x"]),
            Table(q, primary_key=["x"], nullable=["y"]),
            Table(q, primary_key=["x"], nullable=["x", "z"]),
            Table(q, primary_key=["x"], non_nullable=["y"]),
            Table(q, primary_key=["x"], non_nullable=["x", "z"]),
            Table(q, primary_key=["x"], nullable=["x", "y", "z"], non_nullable=[]),
            Table(q, primary_key=["x"], nullable=["x", "z"], non_nullable=["y"]),
            Table(q, primary_key=["x"], nullable=["y"], non_nullable=["x", "z"]),
            Table(q, primary_key=["x"], nullable=[], non_nullable=["x", "y", "z"]),
        ]
        return tables

    with Flow() as f:
        with Stage("stage_1"):
            lazy_tables = lazy_task_1()
            _check_nullable_task(lazy_tables)
    return f, lazy_tables


def _test_nullable_get_flow():
    @materialize
    def task_1():
        df = pd.DataFrame({"x": [1], "y": [2], "z": [3]})
        tables = [
            Table(df),
            Table(df, nullable=["y"]),
            Table(df, nullable=["x", "z"]),
            Table(df, non_nullable=["y"]),
            Table(df, non_nullable=["x", "z"]),
            Table(df, nullable=["x", "y", "z"], non_nullable=[]),
            Table(df, nullable=["x", "z"], non_nullable=["y"]),
            Table(df, nullable=["y"], non_nullable=["x", "z"]),
            Table(df, nullable=[], non_nullable=["x", "y", "z"]),
            Table(df, primary_key=["x"]),
            Table(df, primary_key=["x"], nullable=["y"]),
            Table(df, primary_key=["x"], nullable=["x", "z"]),
            Table(df, primary_key=["x"], non_nullable=["y"]),
            Table(df, primary_key=["x"], non_nullable=["x", "z"]),
            Table(df, primary_key=["x"], nullable=["x", "y", "z"], non_nullable=[]),
            Table(df, primary_key=["x"], nullable=["x", "z"], non_nullable=["y"]),
            Table(df, primary_key=["x"], nullable=["y"], non_nullable=["x", "z"]),
            Table(df, primary_key=["x"], nullable=[], non_nullable=["x", "y", "z"]),
        ]
        return tables

    with Flow() as f:
        with Stage("stage_1"):
            lazy_tables = task_1()
            _check_nullable_task(lazy_tables)
    return f, lazy_tables


@skip_instances("snowflake", "parquet_backend", "parquet_s3_backend")
@pytest.mark.parametrize("_get_flow", [_test_nullable_get_flow, _test_nullable_lazy_get_flow])
def test_nullable(_get_flow):
    f, tables = _get_flow()

    with StageLockContext():
        result = f.run()
        assert result.successful
        for i in range(18):  # unfortunately, 'for tbl in tables' does not work
            tbl = tables[i]
            assert len(result.get(tbl, as_type=pd.DataFrame)) == 1
            assert result.get(tbl, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(tbl, as_type=pd.DataFrame)["y"][0] == 2
            assert result.get(tbl, as_type=pd.DataFrame)["z"][0] == 3
        # check that there are 9 tables with one primary key column each by checking
        # the sum of all primary key columns
        if ConfigContext.get().store.table_store.engine.dialect.name != "duckdb":
            assert (
                sum(col.primary_key for i in range(18) for col in result.get(tables[i], as_type=sa.Table).original.c)
                == 9
            )


# duckdb does not persist nullable flags over connections; snowflake is slow
@skip_instances("duckdb", "snowflake", "parquet_backend", "parquet_s3_backend")
@pytest.mark.parametrize("_get_flow", [_test_nullable_get_flow, _test_nullable_lazy_get_flow])
def test_nullable_output(_get_flow):
    f, tables = _get_flow()

    def get_nullable(idx):
        return [c.nullable for c in result.get(tables[idx], as_type=sa.Table).c]

    with StageLockContext():
        result = f.run()
        assert result.successful
        # This is dialect specific:
        # assert get_nullable(0) == [True, True, True]
        assert get_nullable(1) == [False, True, False]
        assert get_nullable(2) == [True, False, True]
        assert get_nullable(3) == [True, False, True]
        assert get_nullable(4) == [False, True, False]
        assert get_nullable(5) == [True, True, True]
        assert get_nullable(6) == [True, False, True]
        assert get_nullable(7) == [False, True, False]
        assert get_nullable(8) == [False, False, False]


@pytest.mark.parametrize(
    "nullable, non_nullable, error",
    [
        (["a"], None, ValueError),
        (None, ["a"], ValueError),
        (["a", "x", "y", "z"], None, ValueError),
        (None, ["a", "x", "y", "z"], ValueError),
        (["x", "y", "z"], ["a"], ValueError),
        (["a"], ["x", "y", "z"], ValueError),
        (["x"], ["x"], ValueError),
        (["x"], ["y"], ValueError),
        (["x", "y"], ["y"], ValueError),
        ("x", None, TypeError),
        (None, "x", TypeError),
        (1, None, TypeError),
        (None, 2, TypeError),
    ],
)
def test_nullable_raises(nullable, non_nullable, error):
    cols = sa.literal(1).label("x"), sa.literal(2).label("y"), sa.literal(3).label("z")

    @materialize(lazy=True)
    def lazy_task_1():
        q = sa.select(*cols)
        df = pd.DataFrame({"x": [1], "y": [2], "z": [3]})
        tables = [
            Table(q, nullable=nullable, non_nullable=non_nullable),
            Table(df, nullable=nullable, non_nullable=non_nullable),
            Table(q, primary_key=["x"], nullable=nullable, non_nullable=non_nullable),
            Table(df, primary_key=["x"], nullable=nullable, non_nullable=non_nullable),
        ]
        return tables

    with Flow() as f:
        with Stage("stage_1"):
            lazy_task_1()

    with pytest.raises(error):
        f.run()


@materialize(lazy=True)
def _lazy_task_1():
    return Table(select_as(1, "x"), name="t1", primary_key="x")


@materialize(lazy=True, nout=2)
def _lazy_task_2():
    return Table(select_as(1, "x"), name="t21", indexes=[["x"]]), Table(
        select_as(2, "x"), name="t22", primary_key=["x"]
    )


@materialize(lazy=True, input_type=sa.Table)
def _lazy_join(src1: sa.sql.expression.Alias, src2: sa.sql.expression.Alias):
    query = sa.select(src1.c.x, src2.c.x.label("x2")).select_from(src1.outerjoin(src2, src1.c.x == src2.c.x))
    return Table(query, "t3_%%", indexes=[["x2"], ["x", "x2"]])


@materialize(version="1.0")
def _sql_task_1():
    return Table(select_as(1, "x"), name="t1", primary_key="x")


@materialize(version="1.0", nout=2)
def _sql_task_2():
    return (
        Table(select_as(1, "x"), name="t21", indexes=[["x"]]),
        Table(select_as(2, "x"), name="t22", primary_key=["x"]),
    )


@materialize(version="1.0", input_type=sa.Table)
def _sql_join(src1: sa.sql.expression.Alias, src2: sa.sql.expression.Alias):
    query = sa.select(src1.c.x, src2.c.x.label("x2")).select_from(src1.outerjoin(src2, src1.c.x == src2.c.x))
    return Table(query, "t3_%%", indexes=[["x2"], ["x", "x2"]])


@materialize(version="1.0")
def _eager_task_1():
    df = pd.DataFrame(dict(x=[1]))
    return Table(df, name="t1", primary_key="x")


@materialize(version="1.0", nout=2)
def _eager_task_2():
    df1 = pd.DataFrame(dict(x=[1]))
    df2 = pd.DataFrame(dict(x=[2]))
    return (
        Table(df1, name="t21", indexes=[["x"]]),
        Table(df2, name="t22", primary_key=["x"]),
    )


@materialize(version="1.0", input_type=pd.DataFrame)
def _eager_join(src1: pd.DataFrame, src2: pd.DataFrame):
    src2["x2"] = src2["x"]
    join = src1.merge(src2, on="x", how="left")
    return Table(join, "t3_%%", indexes=[["x2"], ["x", "x2"]])


@pytest.mark.skip
@pytest.mark.parametrize(
    "task_1,task_2,noop,join",
    [
        # communicate with same kind
        (_lazy_task_1, _lazy_task_2, m.noop_lazy, _lazy_join),
        (_sql_task_1, _sql_task_2, m.noop_sql, _sql_join),
        (_eager_task_1, _eager_task_2, m.noop, _eager_join),
        # communicate to sql
        (_lazy_task_1, _lazy_task_2, m.noop_sql, _sql_join),
        (_eager_task_1, _eager_task_2, m.noop_sql, _sql_join),
        # communicate to lazy
        (_sql_task_1, _sql_task_2, m.noop_lazy, _lazy_join),
        (_eager_task_1, _eager_task_2, m.noop_lazy, _lazy_join),
        # communicate to eager
        (_lazy_task_1, _lazy_task_2, m.noop, _eager_join),
        (_sql_task_1, _sql_task_2, m.noop, _eager_join),
    ],
)
def test_task_and_stage_communication(task_1, task_2, noop, join):
    logger = structlog.get_logger("test_task_and_stage_communication")
    with Flow() as f:
        with Stage("stage_1"):
            t1 = task_1()
            t21, t22 = task_2()
            same_stage = noop(t22)
            same_stage2 = join(t1, t21)
        with Stage("stage_2"):
            next_stage = noop(t22)
            next_stage2 = join(t1, t21)
            next_same_stage = join(t1, next_stage)

    for i in range(3):
        with StageLockContext():
            logger.info("## Running flow", iteration=i)
            result = f.run()
            logger.info("#    checking flow", iteration=i)
            assert result.successful
            assert result.get(t1, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(t21, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(t22, as_type=pd.DataFrame)["x"][0] == 2
            assert result.get(same_stage, as_type=pd.DataFrame)["x"][0] == 2
            assert result.get(same_stage2, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(same_stage2, as_type=pd.DataFrame)["x2"][0] == 1
            assert result.get(next_stage, as_type=pd.DataFrame)["x"][0] == 2
            assert result.get(next_stage2, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(next_stage2, as_type=pd.DataFrame)["x2"][0] == 1
            assert result.get(next_same_stage, as_type=pd.DataFrame)["x"][0] == 1
            assert result.get(next_same_stage, as_type=pd.DataFrame)["x2"].isna().all()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, Stage, Table, materialize
from pydiverse.pipedag.backend.table.sql.dialects.ibm_db2 import (
    IBMDB2TableStore,
)
from pydiverse.pipedag.backend.table.sql.dialects.mssql import (
    MSSqlTableStore,
)

# Parameterize all tests in this file with several instance_id configurations
from tests.fixtures.instances import (
    DATABASE_INSTANCES,
    skip_instances,
    with_instances,
)
from tests.util import tasks_library as m

pytestmark = [with_instances(DATABASE_INSTANCES)]


@pytest.mark.parametrize(
    "task, stage_materialization_details",
    [
        (m.simple_table_compressed_one_method, "adaptive_value_compression"),
        (m.simple_table_compressed_two_methods, "adaptive_value_compression"),
        (m.simple_dataframe_compressed_one_method, "adaptive_value_compression"),
        (m.simple_dataframe_compressed_two_methods, "adaptive_value_compression"),
        (m.simple_table_default_compressed, "adaptive_value_compression"),
        (m.simple_dataframe_uncompressed, None),
    ],
)
@with_instances(DATABASE_INSTANCES, "ibm_db2_materialization_details")
@skip_instances("ibm_db2", "parquet_backend", "parquet_s3_backend")
def test_compression(task, stage_materialization_details):
    @materialize(input_type=sa.Table, lazy=False)
    def get_compression_attributes(table: sa.sql.expression.Alias):
        query = f"""
        SELECT COMPRESSION, ROWCOMPMODE FROM SYSCAT.TABLES
        WHERE TABSCHEMA = '{table.original.schema.upper()}'
         AND TABNAME = '{table.original.name.upper()}'
        """
        return Table(sa.text(query), f"compression_attributes_{table.name}")

    with Flow("flow") as f:
        with Stage("stage", materialization_details=stage_materialization_details):
            comp_exp_x, x = task()
            config = ConfigContext.get()
            store = config.store.table_store
            if isinstance(store, IBMDB2TableStore):
                comp_x = get_compression_attributes(x)
                m.assert_table_equal(comp_exp_x, comp_x)

            m.assert_table_equal(x, x)

    for _ in range(3):
        if not isinstance(store, (MSSqlTableStore, IBMDB2TableStore)) and task != m.simple_dataframe_uncompressed:
            with pytest.raises(
                ValueError,
                match="To silence this exception set strict_materialization_details=False",
            ):
                assert f.run().successful
        else:
            assert f.run().successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import os
import pickle
import shutil
from typing import Any

from upath import UPath

from pydiverse.pipedag import Blob, Stage
from pydiverse.pipedag.context import ConfigContext
from pydiverse.pipedag.errors import CacheError
from pydiverse.pipedag.materialize.store import BaseBlobStore
from pydiverse.pipedag.util import normalize_name

__all__ = [
    "FileBlobStore",
]

from pydiverse.pipedag.util.path import is_file_uri


class NoBlobStore(BaseBlobStore):
    """No blob store

    The ``NoBlobStore`` can be used for pipelines where no task outputs any blob.
    """

    def init_stage(self, stage: Stage):
        pass

    def commit_stage(self, stage: Stage):
        pass

    def store_blob(self, blob: Blob):
        raise AttributeError(
            f"Using NoBlobStore is only allowed for pipelines where no task outputs "
            f"any blob. See {blob.stage.name}.{blob.name} ({type(blob.obj)})"
        )

    def copy_blob_to_transaction(self, blob: Blob):
        raise AttributeError(
            f"Using NoBlobStore is only allowed for pipelines where no task outputs "
            f"any blob. See {blob.stage.name}.{blob.name} ({type(blob.obj)})"
        )

    def delete_blob_from_transaction(self, blob: Blob):
        raise AttributeError(
            f"Using NoBlobStore is only allowed for pipelines where no task outputs "
            f"any blob. See {blob.stage.name}.{blob.name} ({type(blob.obj)})"
        )

    def retrieve_blob(self, blob: Blob):
        raise AttributeError(
            f"Using NoBlobStore is only allowed for pipelines where no task outputs "
            f"any blob. See {blob.stage.name}.{blob.name} ({type(blob.obj)})"
        )


class FileBlobStore(BaseBlobStore):
    """File based blob store

    The ``FileBlobStore`` stores blobs in a folder structure on a file system.
    In the base directory there will be two folders for every stage, one
    for the base and one for the transaction stage. Inside those folders the
    blobs will be stored as pickled files:
    ``base_path/instance_id/STAGE_NAME/BLOB_NAME.pkl``.

    To commit a stage, the only thing that has to be done is to rename
    the appropriate folders.
    """

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        instance_id = normalize_name(ConfigContext.get().instance_id)
        base_path = UPath(config["base_path"]) / instance_id
        return cls(base_path)

    def __init__(self, base_path: str | UPath):
        self.base_path = UPath(base_path).absolute()
        if is_file_uri(self.base_path):
            os.makedirs(self.base_path, exist_ok=True)

    def init_stage(self, stage: Stage):
        stage_path = self.get_stage_path(stage.name)
        transaction_path = self.get_stage_path(stage.transaction_name)

        if is_file_uri(stage_path):
            try:
                os.mkdir(stage_path)
            except FileExistsError:
                pass

        if is_file_uri(transaction_path):
            try:
                os.mkdir(transaction_path)
            except FileExistsError:
                shutil.rmtree(transaction_path)
                os.mkdir(transaction_path)
        else:
            # clear bucket key prefix similar to removing a directory
            if transaction_path.exists():
                transaction_path.fs.rmdir(transaction_path)

    def commit_stage(self, stage: Stage):
        stage_path = self.get_stage_path(stage.name)
        transaction_path = self.get_stage_path(stage.transaction_name)
        tmp_path = self.get_stage_path(stage.name + "__swap")

        if not is_file_uri(transaction_path):
            # TODO: think about switching to odd/even transaction schemas
            # to avoid renaming all files in a schema.
            # Another alternative: store all files written to transaction
            raise NotImplementedError("Using Bucket as base path is currently not implemented.")
        os.rename(transaction_path, tmp_path)
        os.rename(stage_path, transaction_path)
        os.rename(tmp_path, stage_path)
        shutil.rmtree(transaction_path)

    def store_blob(self, blob: Blob):
        with open(self.get_blob_path(blob.stage.transaction_name, blob.name), "wb") as f:
            pickle.dump(blob.obj, f, pickle.HIGHEST_PROTOCOL)

    def copy_blob_to_transaction(self, blob: Blob):
        from_path = self.get_blob_path(blob.stage.name, blob.name)
        to_path = self.get_blob_path(blob.stage.transaction_name, blob.name)
        if is_file_uri(from_path):
            try:
                shutil.copy2(
                    from_path,
                    to_path,
                )
            except FileNotFoundError:
                raise CacheError(
                    f"Can't copy blob '{blob.name}' (stage: '{blob.stage.name}')"
                    " to working transaction because no such blob exists."
                ) from None
        else:
            from_path.fs.copy(from_path, to_path, recursive=True)

    def delete_blob_from_transaction(self, blob: Blob):
        self.get_blob_path(blob.stage.transaction_name, blob.name).unlink(missing_ok=True)

    def retrieve_blob(self, blob: Blob):
        stage = blob.stage

        with open(self.get_blob_path(stage.current_name, blob.name), "rb") as f:
            return pickle.load(f)

    def get_stage_path(self, stage_name: str) -> UPath:
        return self.base_path / stage_name

    def get_blob_path(self, stage_name: str, blob_name: str) -> UPath:
        return self.base_path / stage_name / (blob_name + ".pkl")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import signal
from contextlib import contextmanager


@contextmanager
def timeout(seconds=1, error_message="timeout reached"):
    """
    Context manager to raise a TimeoutError after a specified number of seconds.
    """

    def alarm_handler(signum, frame):
        _ = signum, frame
        raise TimeoutError(error_message)

    signal.signal(signal.SIGALRM, alarm_handler)
    signal.alarm(seconds)

    try:
        yield
    finally:
        signal.alarm(0)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy

import sqlalchemy as sa

from pydiverse.pipedag.context import ConfigContext
from pydiverse.pipedag.materialize.store import BaseTableStore


def select_as(value, as_):
    return sa.select(sa.literal(value).label(as_))


def sql_table_expr(cols: dict):
    num_values = {len(vals) for vals in cols.values()}
    assert len(num_values) == 1

    queries = []
    num_values = num_values.pop()
    for i in range(num_values):
        literals = []
        for col, vals in cols.items():
            literals.append(sa.literal(vals[i]).label(col))

        queries.append(sa.select(*literals))

    return sa.union_all(*queries)


def get_config_with_table_store(base_cfg: ConfigContext, table_store_class: type[BaseTableStore]):
    instance = base_cfg.instance_name
    flow = base_cfg.flow_name
    cfg = ConfigContext.new(copy.deepcopy(base_cfg._config_dict), base_cfg.pipedag_name, flow, instance)
    cfg._config_dict["table_store"]["class"] = table_store_class
    # this actually instantiates the table store
    table_store = cfg.store.table_store
    assert type(table_store) == table_store_class  # noqa: E721
    return cfg



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pathlib import Path

import filelock
import pandas as pd
import pytest
import sqlalchemy as sa

from pydiverse.pipedag import ConfigContext, Flow, RawSql, Stage, Table, materialize
from pydiverse.pipedag.backend.table.sql.ddl import (
    CreateSchema,
    CreateTableAsSelect,
    DropTable,
)
from pydiverse.pipedag.container import ExternalTableReference, Schema
from tests.fixtures.instances import with_instances
from tests.util.sql import sql_table_expr
from tests.util.tasks_library import (
    assert_table_equal,
    noop_sql,
    simple_dataframe,
    simple_lazy_table,
)


@with_instances("ibm_db2")
def test_db2_nicknames():
    lock_path = Path(__file__).parent / "scripts" / "lock"

    @materialize(input_type=sa.Table)
    def create_nicknames(table: sa.sql.expression.Alias):
        script_path = Path(__file__).parent / "scripts" / "simple_nicknames.sql"
        simple_nicknames = Path(script_path).read_text(encoding="utf-8")
        simple_nicknames = simple_nicknames.replace("{{out_schema}}", str(table.original.schema))
        simple_nicknames = simple_nicknames.replace("{{out_table}}", str(table.original.name))

        return RawSql(simple_nicknames, "create_nicknames", separator="|")

    with Flow("f") as f:
        with Stage("stage"):
            x = simple_dataframe()
            nicknames = create_nicknames(x)
            _ = nicknames

    with filelock.FileLock(lock_path):
        # We run three times to ensure that the nicknames created in the first run
        # have to be dropped, since the same schema is reused.
        assert f.run().successful
        assert f.run().successful
        assert f.run().successful


@with_instances("ibm_db2")  # only one instance to avoid parallel DRDA WRAPPER creation
def test_db2_table_reference_nicknames():
    lock_path = Path(__file__).parent / "scripts" / "lock"

    @materialize(nout=2)
    def create_external_nicknames():
        table_store = ConfigContext.get().store.table_store
        schema = Schema("user_controlled_schema", prefix="", suffix="")
        table_name = "external_table_for_nickname"
        table_store.execute(CreateSchema(schema, if_not_exists=True))
        table_store.execute(DropTable(table_name, schema, if_exists=True))
        query = sql_table_expr({"col": [0, 1, 2, 3]})
        table_store.execute(
            CreateTableAsSelect(
                table_name,
                schema,
                query,
            )
        )
        script_path = Path(__file__).parent / "scripts" / "simple_nicknames.sql"
        simple_nicknames = Path(script_path).read_text()
        simple_nicknames = simple_nicknames.replace("{{out_schema}}", schema.get())
        simple_nicknames = simple_nicknames.replace("{{out_table}}", table_name)

        table_store.execute_raw_sql(RawSql(simple_nicknames, "create_external_nicknames", separator="|"))

        return Table(ExternalTableReference("nick1", schema=schema.get(), shared_lock_allowed=False)), Table(
            ExternalTableReference("nick2", schema=schema.get(), shared_lock_allowed=True)
        )

    with Flow("f") as f:
        with Stage("stage"):
            nick_1_ref, nick_2_ref = create_external_nicknames()
            nick_1_ref_noop = noop_sql(nick_1_ref)
            nick_2_ref_noop = noop_sql(nick_2_ref)
            assert_table_equal(nick_1_ref, nick_1_ref_noop)
            assert_table_equal(nick_2_ref, nick_2_ref_noop)

    with filelock.FileLock(lock_path):
        # We run three times to ensure that the nicknames created in the first run
        # have to be dropped, since the same schema is reused.
        assert f.run().successful
        assert f.run().successful
        assert f.run().successful


@with_instances("ibm_db2_materialization_details")
@pytest.mark.parametrize("task", [simple_dataframe, simple_lazy_table])
def test_db2_table_spaces(task):
    @materialize()
    def create_table_spaces():
        script_path = Path(__file__).parent / "scripts" / "simple_table_spaces.sql"
        simple_table_spaces = Path(script_path).read_text(encoding="utf-8")
        return RawSql(simple_table_spaces, "create_table_spaces", separator="|")

    @materialize(input_type=sa.Table, lazy=False)
    def get_actual_table_space_attributes(table: sa.sql.expression.Alias):
        query = f"""
            SELECT TBSPACE, INDEX_TBSPACE, LONG_TBSPACE FROM SYSCAT.TABLES
             WHERE TABSCHEMA = '{table.original.schema.upper()}'
             AND TABNAME = '{table.original.name.upper()}'
        """
        return Table(sa.text(query), f"tbspace_attributes_{table.name}")

    @materialize(version="1.0")
    def get_expected_table_space_attributes():
        return Table(
            pd.DataFrame({"TBSPACE": ["S1"], "INDEX_TBSPACE": ["S2"], "LONG_TBSPACE": ["S3"]}),
            name="tbspace_attributes",
        )

    with Flow("f") as f:
        with Stage("stage", materialization_details="table_space"):
            create_table_spaces()
            x = task()
            x_attr = get_actual_table_space_attributes(x)
            attrs = get_expected_table_space_attributes()
            assert_table_equal(attrs, x_attr)

    for _ in range(3):
        assert f.run().successful



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import csv
from dataclasses import dataclass
from io import BytesIO, StringIO
from typing import Any

import pandas as pd
import sqlalchemy as sa

from pydiverse.common import Dtype
from pydiverse.pipedag.backend.table.sql import hooks
from pydiverse.pipedag.backend.table.sql.ddl import (
    ChangeTableLogged,
    LockSourceTable,
    LockTable,
)
from pydiverse.pipedag.backend.table.sql.sql import SQLTableStore
from pydiverse.pipedag.container import Schema, Table
from pydiverse.pipedag.materialize.details import (
    BaseMaterializationDetails,
    resolve_materialization_details_label,
)

try:
    import polars as pl
except ImportError:
    pl = None

try:
    import pyarrow as pa
except ImportError:
    pa = None


@dataclass(frozen=True)
class PostgresMaterializationDetails(BaseMaterializationDetails):
    """
    Materialization details specific to PostgreSQL.

    :param unlogged: Whether to use `unlogged`_ tables or not.
        This reduces safety in case of a crash or unclean shutdown, but can
        significantly increase write performance.

    .. _unlogged:
        https://www.postgresql.org/docs/9.5/sql-createtable.html
        #SQL-CREATETABLE-UNLOGGED

    Example
    -------

    .. code-block:: yaml

        materialization_details:
            __any__:
                unlogged: true
            my_label:
                unlogged: false

    For more general information on materialization details, see
    ``materialization_details`` parameter of
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`.
    """

    def __post_init__(self):
        assert isinstance(self.unlogged, bool)

    unlogged: bool = False


class PostgresTableStore(SQLTableStore):
    """
    SQLTableStore that supports `PostgreSQL <https://postgresql.org>`_.

    Takes the same arguments as
    :py:class:`SQLTableStore <pydiverse.pipedag.backend.table.SQLTableStore>`.

    Supports the :py:class:`PostgresMaterializationDetails\
    <pydiverse.pipedag.backend.table.sql.dialects.postgres.PostgresMaterializationDetails>`
    materialization details.
    """

    _dialect_name = "postgresql"

    def __init__(self, *args, use_adbc: bool = True, **kwargs):
        self.use_adbc = use_adbc
        super().__init__(*args, **kwargs)

    def _init_database(self):
        self._init_database_with_database("postgres", {"isolation_level": "AUTOCOMMIT"})

    def get_unlogged(self, materialization_details_label: str | None) -> bool:
        return PostgresMaterializationDetails.get_attribute_from_dict(
            self.materialization_details,
            materialization_details_label,
            self.default_materialization_details,
            "unlogged",
            self.strict_materialization_details,
            self.logger,
        )

    def lock_table(self, table: Table | str, schema: Schema | str, conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock a table before writing to it.
        """
        stmt = LockTable(table.name if isinstance(table, Table) else table, schema)
        if conn is not None:
            self.execute(stmt, conn=conn)
        return [stmt]

    def lock_source_table(self, table: Table | str, schema: Schema | str, conn: Any = None) -> list:
        """
        For some dialects, it might be beneficial to lock source tables before reading.
        """
        stmt = LockSourceTable(table.name if isinstance(table, Table) else table, schema)
        if conn is not None:
            self.execute(stmt, conn=conn)
        return [stmt]

    def check_materialization_details_supported(self, label: str | None) -> None:
        _ = label
        return

    def _set_materialization_details(self, materialization_details: dict[str, dict[str | list[str]]] | None) -> None:
        self.materialization_details = PostgresMaterializationDetails.create_materialization_details_dict(
            materialization_details,
            self.strict_materialization_details,
            self.default_materialization_details,
            self.logger,
        )


@PostgresTableStore.register_table()
class SQLAlchemyTableHook(hooks.SQLAlchemyTableHook):
    pass  # postges is our reference dialect


@PostgresTableStore.register_table(pd)
class PandasTableHook(hooks.PandasTableHook):
    @classmethod
    def _execute_materialize(
        cls,
        table: Table[pd.DataFrame],
        store: PostgresTableStore,
        schema: Schema,
        dtypes: dict[str, Dtype],
    ):
        df = table.obj
        dtypes = cls._get_dialect_dtypes(dtypes, table)

        # Create empty table
        cls._dialect_create_empty_table(store, table, schema, dtypes)
        store.add_indexes_and_set_nullable(table, schema, on_empty_table=True, table_cols=df.columns)

        if store.get_unlogged(resolve_materialization_details_label(table)):
            store.execute(ChangeTableLogged(table.name, schema, False))

        # COPY data
        # TODO: For python 3.12, there is csv.QUOTE_STRINGS
        #       This would make everything a bit safer, because then we could
        #       represent the string "\\N" (backslash + capital n).
        s_buf = StringIO()
        df.to_csv(
            s_buf,
            na_rep="\\N",
            header=False,
            index=False,
            quoting=csv.QUOTE_MINIMAL,
        )
        s_buf.seek(0)

        engine = store.engine
        table_name = engine.dialect.identifier_preparer.quote(table.name)
        schema_name = engine.dialect.identifier_preparer.format_schema(schema.get())

        dbapi_conn = engine.raw_connection()
        try:
            with dbapi_conn.cursor() as cur:
                sql = f"COPY {schema_name}.{table_name} FROM STDIN WITH (FORMAT CSV, NULL '\\N')"
                store.logger.info("Executing bulk load", query=sql)
                cur.copy_expert(sql=sql, file=s_buf)
            dbapi_conn.commit()
        finally:
            dbapi_conn.close()
        store.add_indexes_and_set_nullable(
            table,
            schema,
            on_empty_table=False,
        )


@PostgresTableStore.register_table(pl)
class PolarsTableHook(hooks.PolarsTableHook):
    @classmethod
    def dialect_supports_connectorx(cls):
        # ConnectorX (used by Polars read_database_uri) supports PostgreSQL.
        return True

    @classmethod
    def upload_table(
        cls,
        table: Table[pl.DataFrame],
        schema: Schema,
        dtypes: dict[str, sa.types.TypeEngine],
        store: SQLTableStore,
        early: bool,
    ):
        if not early:
            cls._dialect_create_empty_table(store, table, schema, dtypes)
        df = table.obj  # type: pl.DataFrame

        # COPY data
        # TODO: For python 3.12, there is csv.QUOTE_STRINGS
        #       This would make everything a bit safer, because then we could
        #       represent the string "\\N" (backslash + capital n).
        s_buf = BytesIO()
        df.write_csv(
            s_buf,
            null_value="\\N",
            include_header=False,
            quote_style="necessary",
        )
        s_buf.seek(0)

        engine = store.engine
        table_name = engine.dialect.identifier_preparer.quote(table.name)
        schema_name = engine.dialect.identifier_preparer.format_schema(schema.get())

        dbapi_conn = engine.raw_connection()
        try:
            with dbapi_conn.cursor() as cur:
                sql = f"COPY {schema_name}.{table_name} FROM STDIN WITH (FORMAT CSV, NULL '\\N')"
                store.logger.info("Executing bulk load", query=sql)
                cur.copy_expert(sql=sql, file=s_buf)
            dbapi_conn.commit()
        finally:
            dbapi_conn.close()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import tempfile

from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import Flow, GroupNode, Stage, VisualizationStyle, materialize
from pydiverse.pipedag.core.config import create_basic_pipedag_config


@materialize
def any_task():
    return 1


@materialize
def task_within_group():
    return 2


@materialize
def task_within_group2(input1: int):
    return input1 + 1


def main():
    with tempfile.TemporaryDirectory() as temp_dir:
        cfg = create_basic_pipedag_config(
            f"duckdb:///{temp_dir}/db.duckdb",
            disable_stage_locking=True,  # This is special for duckdb
            # Attention: stage and task names might be sent to the
            #   following URL. You can self-host kroki if you like:
            #   https://docs.kroki.io/kroki/setup/install/
            #   You need to install optional dependency 'pydot' for any visualization
            #   URL to appear.
            kroki_url="https://kroki.io",
        ).get("default")
        with cfg:
            with Flow() as flow:
                with Stage("stage1"):
                    _ = any_task()
                    with GroupNode(
                        "group1",
                        ordering_barrier=True,
                        style=VisualizationStyle(hide_content=True, box_color_always="#ccccff"),
                    ):
                        task1 = task_within_group()
                        _ = task_within_group2(task1)
                    _ = any_task()

            # Run flow
            result = flow.run()
            assert result.successful

            # you can also visualize the flow explicitly:
            # kroki_url = result.visualize_url()
            # result.visualize()


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish
    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

# register all SQLTableStore dialects (otherwise, they cannot be selected by DB URL)
from .duckdb import DuckDBTableStore
from .ibm_db2 import IBMDB2TableStore
from .mssql import MSSqlTableStore
from .postgres import PostgresTableStore
from .snowflake import SnowflakeTableStore

__all__ = [
    "DuckDBTableStore",
    "IBMDB2TableStore",
    "MSSqlTableStore",
    "PostgresTableStore",
    "SnowflakeTableStore",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from .blob import FileBlobStore, NoBlobStore
from .table.parquet import ParquetTableStore
from .table.sql import SQLTableStore

__all__ = [
    "ParquetTableStore",
    "SQLTableStore",
    "FileBlobStore",
    "NoBlobStore",
]



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import copy
import unittest.mock

from pydiverse.pipedag.core.task import Task, TaskGetItem
from pydiverse.pipedag.materialize.materializing_task import MaterializingTask


class PipedagMock:
    def __init__(self, mock: unittest.mock.Mock):
        self.mock = mock
        self._last_call_count = mock.call_count

    def reset_call_count(self):
        self._last_call_count = self.mock.call_count

    def _calls_since_last_time(self):
        delta = self.mock.call_count - self._last_call_count
        self.reset_call_count()
        return delta

    def _assert_call_count(self, n):
        __tracebackhide__ = True
        m = self._calls_since_last_time()
        if n == m:
            return
        name = self.mock.mock.__dict__["_mock_name"]
        msg = (
            f"Expected function '{name}' to have been called {n} times, but it has"
            f" been called {m} times ({self.mock.call_count} times in total)."
        )
        raise AssertionError(msg)

    def assert_not_called(self):
        __tracebackhide__ = True
        self._assert_call_count(0)

    def assert_called_once(self):
        __tracebackhide__ = True
        self._assert_call_count(1)

    def assert_called(self, times):
        __tracebackhide__ = True
        self._assert_call_count(times)


def spy_task(mocker, task) -> PipedagMock:
    if isinstance(task, TaskGetItem):
        task = task.task
    if isinstance(task, MaterializingTask):
        task.fn = copy.copy(task.fn)
        spy = mocker.spy(task.fn, "fn")
    elif isinstance(task, Task):
        task_fn = task.fn

        def fn(*args, **kwargs):
            return task_fn(*args, **kwargs)

        task.fn = fn
        spy = mocker.spy(task, "fn")
    else:
        raise TypeError("Expected object of type Task or TaskGetItem")

    spy.mock.__dict__["_mock_name"] = task.name
    return PipedagMock(spy)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause
import os

import dataframely as dy
import pandas as pd
import polars as pl
import sqlalchemy as sa

import pydiverse.colspec as cs
import pydiverse.transform as pdt
from pydiverse.common.util.structlog import setup_logging
from pydiverse.pipedag import AUTO_VERSION, Flow, Stage, Table, materialize
from pydiverse.pipedag.context import StageLockContext
from pydiverse.pipedag.util.testing_s3 import initialize_test_s3_bucket
from pydiverse.transform.extended import left_join


@materialize(lazy=True)
def lazy_task_1():
    return Table(
        sa.select(
            sa.literal(1).label("x"),
            sa.literal(2).label("y"),
        ),
        name="lazy_1",
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_2(input1: sa.Alias, input2: sa.Alias):
    query = sa.select(
        (input1.c.x * 5).label("x5"),
        input2.c.a,
    ).select_from(input1.outerjoin(input2, input2.c.x == input1.c.x))

    return Table(query, name="task_2_out", primary_key=["a"])


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_3(input1: sa.Alias):
    return sa.text(f"SELECT * FROM {input1.original.schema}.{input1.original.name}")


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_4(input1: sa.Alias):
    return sa.text(f"SELECT * FROM {input1.original.schema}.{input1.original.name}")


@materialize(nout=2, version="1.0.0")
def eager_inputs():
    dfA = pd.DataFrame(
        {
            "a": [0, 1, 2, 4],
            "b": [9, 8, 7, 6],
        }
    )
    dfB = pd.DataFrame(
        {
            "a": [2, 1, 0, 1],
            "x": [1, 1, 2, 2],
        }
    )
    return Table(dfA, "dfA"), Table(dfB, "dfB_%%")


@materialize(version="1.0.0", input_type=pd.DataFrame)
def eager_task(tbl1: pd.DataFrame, tbl2: pd.DataFrame):
    return tbl1.merge(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_polars(tbl1: pl.LazyFrame, tbl2: pl.LazyFrame):
    return tbl1.join(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pdt.Polars)
def eager_task_pdt(tbl1: pdt.Table, tbl2: pdt.Table):
    return tbl1 >> left_join(tbl2, tbl1.x == tbl2.x) >> pdt.alias("eager_task_pdt_out")


@materialize(lazy=True, input_type=pdt.SqlAlchemy)
def lazy_task_pdt(tbl1: pdt.Table, tbl2: pdt.Table):
    # pydiverse.transform syntax works both on Polars and SQL/duckdb backends.
    return tbl1 >> left_join(tbl2, tbl1.x == tbl2.x) >> pdt.alias("lazy_task_pdt_out")


class Tbl1Schema(dy.Schema):
    x = dy.Int32()
    y = dy.Int32()


class OutputSchema(Tbl1Schema):
    a = dy.Int64()


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_dataframely(tbl1: dy.LazyFrame[Tbl1Schema], tbl2: pl.LazyFrame) -> OutputSchema:
    # Pipedag automatically calls Tbl1Schema.cast() and OutputSchema.validate().
    return tbl1.join(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pdt.Polars)
def eager_task_dataframely_pdt(tbl1: Tbl1Schema, tbl2: pdt.Table) -> OutputSchema:
    # Pipedag automatically calls Tbl1Schema.cast() and OutputSchema.validate().
    return (
        tbl1
        >> left_join(tbl2, tbl1.x == tbl2.x)
        >> pdt.rename({f"a_{tbl2._ast.name}": "a"})
        >> pdt.alias("eager_task_dataframely_pdt_out")
    )


class Tbl1ColSpec(cs.ColSpec):
    x = cs.Int32()
    y = cs.Int32()


class OutputColSpec(Tbl1ColSpec):
    a = cs.Int64()


@materialize(version=AUTO_VERSION, input_type=pl.LazyFrame)
def eager_task_colspec(tbl1: Tbl1ColSpec, tbl2: pl.LazyFrame) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast_polars() and OutputColSpec.validate_polars()
    # which call dataframely in the background. ColSpec classes can also be used for SQL validations.
    return tbl1.join(tbl2, on="x")


@materialize(version=AUTO_VERSION, input_type=pdt.Polars)
def eager_task_colspec_pdt(tbl1: Tbl1ColSpec, tbl2: pdt.Table) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast_polars() and OutputColSpec.validate_polars()
    # which call dataframely in the background. ColSpec classes can also be used for SQL validations.
    return (
        tbl1
        >> left_join(tbl2, tbl1.x == tbl2.x)
        >> pdt.rename({f"a_{tbl2._ast.name}": "a"})
        >> pdt.alias("eager_task_colspec_pdt_out")
    )


@materialize(lazy=True, input_type=sa.Table)
def lazy_task_colspec(tbl1: Tbl1ColSpec, tbl2: sa.Alias) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast() and OutputColSpec.validate()
    return Table(
        sa.select(tbl1, tbl2.c.a).select_from(tbl1.outerjoin(tbl2, tbl1.c.x == tbl2.c.x)), name="lazy_task_colspec"
    )


@materialize(lazy=True, input_type=pdt.SqlAlchemy)
def lazy_task_colspec_pdt(tbl1: Tbl1ColSpec, tbl2: pdt.Table) -> OutputColSpec:
    # Pipedag automatically calls Tbl1ColSpec.cast() and OutputColSpec.validate()
    return (
        tbl1
        >> left_join(tbl2, tbl1.x == tbl2.x)
        >> pdt.rename({f"a_{tbl2._ast.name}": "a"})
        >> pdt.alias("lazy_task_colspec_pdt_out")
    )


def main():
    with Flow() as f:
        with Stage("stage_1"):
            lazy_1 = lazy_task_1()
            a, b = eager_inputs()

        with Stage("stage_2"):
            lazy_2 = lazy_task_2(lazy_1, b)
            lazy_3 = lazy_task_3(lazy_2)
            # Do the same operation with various syntax alternatives.
            # This is also a demonstration how pipedag interacts with \
            # dataframely, pydiverse.colspec, and pydiverse.transform.
            out = eager_task(lazy_1, b)
            out2 = eager_task_polars(lazy_1, b)
            out3 = eager_task_pdt(lazy_1, b)
            out4 = lazy_task_pdt(lazy_1, b)
            out5 = eager_task_dataframely(lazy_1, b)
            out6 = eager_task_dataframely_pdt(lazy_1, b)
            out7 = eager_task_colspec(lazy_1, b)
            out8 = eager_task_colspec_pdt(lazy_1, b)
            out9 = lazy_task_colspec(lazy_1, b)
            out10 = lazy_task_colspec_pdt(lazy_1, b)

        with Stage("stage_3"):
            lazy_4 = lazy_task_4(lazy_2)
        _ = lazy_3, lazy_4, out, out2, out3, out4, out5, out6, out7, out8, out9, out10  # unused terminal output tables

    # Run flow
    result = f.run()
    assert result.successful

    # Run in a different way for testing
    with StageLockContext():
        result = f.run()
        assert result.successful
        assert result.get(lazy_1, as_type=pd.DataFrame)["x"][0] == 1


if __name__ == "__main__":
    setup_logging()  # you can setup the logging and/or structlog libraries as you wish

    # You can run this pipeline with:
    # ```shell
    # pixi run python run_pipeline.py
    # ```

    # don't forget to run docker-compose in separate shell to launch minio container:
    # ```shell
    # pixi run docker-compose up
    # ```

    # for testing, we simply create a test bucket in minio container if it does not exist
    os.environ["AWS_ACCESS_KEY_ID"] = "minioadmin"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "minioadmin"
    os.environ["AWS_ENDPOINT_URL"] = "http://localhost:9000"
    initialize_test_s3_bucket()
    # initialize_test_s3_bucket(test_bucket="pipedag-test-bucket")

    main()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.pipedag.backend.table.sql.ddl import insert_into_in_query


def test_insert_into():
    test_pairs = {
        "Select 1": "Select 1 INTO a.b",
        "Select 1 as _from": "Select 1 as _from INTO a.b",
        "Select 1 as afrom": "Select 1 as afrom INTO a.b",
        "Select 1 WHERE TRUE": "Select 1 INTO a.b WHERE TRUE",
        "Select 1 GROUP\nBY x": "Select 1 INTO a.b GROUP\nBY x",
        "Select 1 FROM A GROUP BY x": "Select 1 INTO a.b FROM A GROUP BY x",
        "Select 1 UNION ALL SELECT 2": "Select 1 INTO a.b UNION ALL SELECT 2",
        "Select 1 From X": "Select 1 INTO a.b From X",
        "Select (SELECT 1 FROM Y) From X": "Select (SELECT 1 FROM Y) INTO a.b From X",
        "Select (SELECT (SELECT 1 FROM Z) FROM Y) From X": ("Select (SELECT (SELECT 1 FROM Z) FROM Y) INTO a.b From X"),
        "Select a.[from] from a": "Select a.[from] INTO a.b from a",
        "Select a.[ from ] from a": "Select a.[ from ] INTO a.b from a",
        'Select "from" from a': 'Select "from" INTO a.b from a',
    }
    for raw_query, expected_query in test_pairs.items():
        res = insert_into_in_query(raw_query, "a", "b")
        assert res == expected_query



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import random

import structlog

from pydiverse.common.util.hashing import stable_hash
from pydiverse.pipedag import ConfigContext, Table
from pydiverse.pipedag.backend import SQLTableStore
from pydiverse.pipedag.container import Schema
from pydiverse.pipedag.context import TaskContext
from pydiverse.pipedag.materialize.materializing_task import MaterializingTask
from pydiverse.pipedag.materialize.store import mangle_table_name


def materialize_table(
    table: Table,
    config_context: ConfigContext | None = None,
    schema: Schema | str | None = None,
    debug_suffix: str | None = None,
    flag_task_debug_tainted: bool = True,
    keep_table_name: bool = True,
    drop_if_exists: bool = True,
):
    """
    This function allows the user to materialize a table ad-hoc
    whenever the TaskContext is defined.
    This can be useful during interactive debugging.
    If table.name is not given, then it is derived from the task name and a suffix.
    The suffix is either the debug_suffix (if given) or a random suffix.
    If the table name ends in %%, the %% are also replaced by a suffix.

    :param table: The table to be materialized.
    :param config_context: The config context to be used for materialization.
        If None, the current config context is used. Default: None
    :param schema: The schema to be used for writing table. If None, the
        schema is derived from the table's stage. Default: None
    :param debug_suffix: A suffix to be appended to the table name
        for debugging purposes. Default: None
    :param flag_task_debug_tainted: Whether to flag the task as tainted
        by this debug materialization. The flag will cause an exception if execution is
        continued. Default: True
    :param keep_table_name: if False, the table.name will be equal to the debug name
        after return. Otherwise, the original name of the table is preserved.
        Default: True
    :param drop_if_exists: If True, try to drop the table (if exists) before recreating.
        This is only supported for SQL table stores.
    """
    if config_context is None:
        config_context = ConfigContext.get()
    table_store = config_context.store.table_store

    task: MaterializingTask | None = None
    try:
        task_context = TaskContext.get()
        task = task_context.task  # type: ignore
        task_name = task.name

        table.stage = task.stage
        if schema is None:
            schema = table_store.get_schema(table.stage.transaction_name)
    except LookupError:
        task_name = None
        # LookupError happens if no TaskContext is open
        if schema is None:
            raise ValueError(
                "Parameter schema must be provided if task is not called by normal pipedag orchestration."
            ) from None

    suffix = stable_hash(str(random.randbytes(8))) + "_0000" if debug_suffix is None else ""
    old_table_name = table.name
    table.name = mangle_table_name(table.name, task_name, suffix)
    if debug_suffix is not None:
        table.name += debug_suffix

    if flag_task_debug_tainted and task is not None:
        task.debug_tainted = True

    if drop_if_exists:
        if isinstance(table_store, SQLTableStore):
            table_store.delete_table_from_transaction(table, schema=schema)
        else:
            logger = structlog.get_logger(logger_name="Debug materialize_table")
            logger.warning("drop_if_exists not supported for non SQLTableStore table stores.")
    if task is None:
        hook = table_store.get_m_table_hook(table)
        schema_name = schema.name if table_store.get_schema(schema.name).get() == schema.get() else schema.get()
        if table_store.get_schema(schema_name).get() != schema.get():
            raise ValueError(
                "Schema prefix and postfix must match prefix and postfix of provided "
                "config_context: "
                f"{table_store.get_schema(schema_name).get()} != {schema.get()}"
            )
        hook.materialize(table_store, table, schema_name, without_config_context=True)
    else:
        table_store.store_table(table, task)

    new_table_name = table.name
    if keep_table_name:
        table.name = old_table_name
    return new_table_name



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import os
import warnings
from pathlib import Path
from typing import Any

from pydiverse.common.util import requires
from pydiverse.pipedag import ConfigContext, Stage
from pydiverse.pipedag.backend.lock.base import BaseLockManager, Lockable, LockState
from pydiverse.pipedag.errors import LockError
from pydiverse.pipedag.util import normalize_name

try:
    import filelock as fl
except ImportError as e:
    warnings.warn(str(e), ImportWarning)
    fl = None


@requires(fl, ImportError("FileLockManager requires 'filelock' to be installed."))
class FileLockManager(BaseLockManager):
    """Lock manager that uses lock files

    For details on how exactly the file locking is implemented, check out the
    `filelock documentation`_.

    :param base_path:
        A path to a folder where the lock files should get stored.
        To differentiate between different instances, the ``instance_id`` will
        automatically be appended to the provided path.

    .. _filelock documentation: https://py-filelock.readthedocs.io/en/latest/index.html
    """

    @classmethod
    def _init_conf_(cls, config: dict[str, Any]):
        instance_id = normalize_name(ConfigContext.get().instance_id)
        base_path = Path(config["base_path"]) / instance_id
        return cls(base_path)

    def __init__(self, base_path: str | Path):
        super().__init__()
        self.base_path = Path(base_path).absolute()
        self.locks: dict[Lockable, fl.BaseFileLock] = {}

        os.makedirs(self.base_path, exist_ok=True)

    @property
    def supports_stage_level_locking(self):
        return True

    def acquire(self, lockable: Lockable):
        if lockable not in self.locks:
            lock_path = self.lock_path(lockable)
            self.locks[lockable] = fl.FileLock(lock_path)

        lock = self.locks[lockable]
        if not lock.is_locked:
            self.logger.info(f"Locking '{lockable}'")
        lock.acquire()
        self.set_lock_state(lockable, LockState.LOCKED)

    def release(self, lockable: Lockable):
        if lockable not in self.locks:
            raise LockError(f"No lock '{lockable}' found.")

        lock = self.locks[lockable]
        lock.release()
        if not lock.is_locked:
            self.logger.info(f"Unlocking '{lockable}'")
            del self.locks[lockable]
            self.set_lock_state(lockable, LockState.UNLOCKED)

    def lock_path(self, lock: Lockable) -> Path:
        if isinstance(lock, Stage):
            return self.base_path / (lock.name + ".lock")
        elif isinstance(lock, str):
            return self.base_path / (lock + ".lock")
        else:
            raise NotImplementedError(f"Can't lock object of type '{type(lock).__name__}'")
